# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""Status bar widgets API."""

from __future__ import annotations

# Standard library import
from typing import TYPE_CHECKING

# Third party imports
import qstylizer.parser
from qtpy import PYSIDE2
from qtpy.QtCore import Qt, QSize, QTimer, Signal
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QHBoxLayout, QLabel, QWidget

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.utils.palette import SpyderPalette
from spyder.utils.qthelpers import create_waitspinner
from spyder.utils.stylesheet import MAC

if TYPE_CHECKING:
    import qstylizer.style.StyleSheet
    from qtpy.QtGui import QCloseEvent, QEnterEvent, QEvent, QMouseEvent


class StatusBarWidget(QWidget, SpyderWidgetMixin):
    """
    Base class for status bar widgets.

    These widgets consist by default of an icon, a label and a spinner,
    which are organized from left to right in that order.

    You can also add any other :class:`QWidget` to this layout by setting the
    :attr:`CUSTOM_WIDGET_CLASS` class attribute. It'll be put between the label
    and the spinner.
    """

    ID: str | None = None
    """
    Unique string identifier name for this widget.
    """

    CUSTOM_WIDGET_CLASS: type[QWidget] | None = None
    """
    A custom widget class object to add to the default layout, or
    ``None`` to not add one.
    """

    INTERACT_ON_CLICK: bool = False
    """
    If ``True``, the user can interact with widget when clicking it
    (for example, to show a menu). If ``False``, the default, the widget
    is not interactive.
    """

    sig_clicked: Signal = Signal()
    """
    Signal emitted when the widget is clicked.
    """

    def __init__(
        self,
        parent: QWidget | None = None,
        show_icon: bool = True,
        show_label: bool = True,
        show_spinner: bool = False,
    ) -> None:
        """
        Base class for status bar widgets.

        These are composed of the following widgets, which are arranged
        in a :class:`QHBoxLayout` from left to right:

        * Icon
        * Label
        * Custom :class:`QWidget`
        * Spinner

        Parameters
        ----------
        parent: QWidget | None, optional
            The parent widget of this one, or ``None`` (default).
        show_icon: bool, optional
            If ``True`` (default), show an icon in the widget. If ``False``,
            don't show an icon.
        show_label: bool, optional
            If ``True`` (default), show a label in the widget. If ``False``,
            don't show a label.
        show_spinner: bool, optional
            If ``True``, show a progress spinner in the widget. If ``False``
            (default), don't show a spinner.

        Notes
        -----
        * To use an icon, you need to redefine the :meth:`get_icon` method.
        * To use a label, you need to call :meth:`set_value`.
        """
        if not PYSIDE2:
            super().__init__(parent, class_parent=parent)
        else:
            QWidget.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=parent)

        self._parent = parent

        self.show_icon = show_icon
        self.show_label = show_label
        self.show_spinner = show_spinner

        self.value = None
        self.label_icon = None
        self.label_value = None
        self.spinner = None
        self.custom_widget = None

        # In case the widget has an associated menu
        self.menu = None

        self._set_layout()
        self._css = self._generate_stylesheet()
        self.setStyleSheet(self._css.toString())

    # ---- Private API
    # -------------------------------------------------------------------------
    def _set_layout(self) -> None:
        """Set layout for default widgets."""
        # Icon
        if self.show_icon:
            self._icon = self.get_icon()
            self._pixmap = None
            self._icon_size = QSize(16, 16)  # Should this be adjustable?
            self.label_icon = QLabel(self)
            self.set_icon()

        # Label
        if self.show_label:
            self.label_value = QLabel(self)
            self.set_value("")
            self.label_value.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        # Custom widget
        if self.CUSTOM_WIDGET_CLASS:
            if not issubclass(self.CUSTOM_WIDGET_CLASS, QWidget):
                raise SpyderAPIError(
                    "Any custom status widget must subclass QWidget!"
                )
            self.custom_widget = self.CUSTOM_WIDGET_CLASS(self._parent)

        # Spinner
        if self.show_spinner:
            self.spinner = create_waitspinner(size=14, parent=self)
            self.spinner.hide()

        # Layout setup
        layout = QHBoxLayout(self)
        layout.setSpacing(0)  # Reduce space between icon and label
        if self.show_icon:
            layout.addWidget(self.label_icon)
        if self.show_label:
            layout.addWidget(self.label_value)
        if self.custom_widget:
            layout.addWidget(self.custom_widget)
        if self.show_spinner:
            layout.addWidget(self.spinner)

        layout.addSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(Qt.AlignVCenter)

        # Setup
        self.update_tooltip()

    def _generate_stylesheet(self) -> qstylizer.style.StyleSheet:
        """Generate the widget's stylesheet."""
        # Remove opacity that comes from QDarkstyle.
        # This work around is necessary because qstylizer doesn't have support
        # for the opacity property.
        initial_css = "QToolTip {opacity: 255;}"
        css = qstylizer.parser.parse(initial_css)

        # Make style match the one set for other tooltips in the app
        css.QToolTip.setValues(
            color=SpyderPalette.COLOR_TEXT_1,
            backgroundColor=SpyderPalette.COLOR_ACCENT_2,
            border="none",
            padding="1px 2px",
        )

        return css

    # ---- Public API
    # -------------------------------------------------------------------------
    def get_icon(self) -> QIcon | None:
        """
        Get the widget's icon.

        Returns
        -------
        QIcon | None
            The widget's icon object, or ``None`` if it doesn't have one.
        """
        return None

    def set_icon(self) -> None:
        """
        Set the icon for the status bar widget.

        Returns
        -------
        None
        """
        if self.label_icon:
            icon = self._icon
            self.label_icon.setVisible(icon is not None)
            if icon is not None and isinstance(icon, QIcon):
                self._pixmap = icon.pixmap(self._icon_size)
                self.label_icon.setPixmap(self._pixmap)

    def set_value(self, value: str) -> None:
        """
        Set formatted text value for the widget.

        Parameters
        ----------
        value : str
            The formatted text value to set.

        Returns
        -------
        None
        """
        if self.label_value:
            self.value = value
            self.label_value.setText(value)

    def get_tooltip(self) -> str:
        """
        Get the widget's tooltip text.

        Returns
        -------
        str
            The widget's tooltip text.
        """
        return ""

    def update_tooltip(self) -> str:
        """
        Update the tooltip for the widget.

        Returns
        -------
        str
            The updated tooltip text.
        """
        tooltip = self.get_tooltip()
        if tooltip:
            if self.label_value:
                self.label_value.setToolTip(tooltip)
            if self.label_icon:
                self.label_icon.setToolTip(tooltip)
            self.setToolTip(tooltip)

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def mousePressEvent(self, event: QMouseEvent) -> None:
        """
        Change the widget's background color when it's clicked.

        Parameters
        ----------
        event : QMouseEvent
            The mouse event that was triggered.

        Returns
        -------
        None
        """
        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_6
            )
            self.setStyleSheet(self._css.toString())

        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """
        Change the widget's background color & emit a signal when it's clicked.

        Parameters
        ----------
        event : QMouseEvent
            The mouse release event.

        Returns
        -------
        None
        """
        super().mouseReleaseEvent(event)

        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                # Mac doesn't correctly restore the background color after
                # clicking on a widget that shows a menu
                backgroundColor=(
                    SpyderPalette.COLOR_BACKGROUND_4
                    if MAC and self.menu
                    else SpyderPalette.COLOR_BACKGROUND_5
                )
            )
            self.setStyleSheet(self._css.toString())

        self.sig_clicked.emit()

    def enterEvent(self, event: QEnterEvent) -> None:
        """
        Change the widget's background color and cursor shape on hover.

        Parameters
        ----------
        event : QEnterEvent
            The mouse hover event.

        Returns
        -------
        None
        """
        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
            )
            self.setStyleSheet(self._css.toString())

            self.setCursor(Qt.PointingHandCursor)

        self.update_tooltip()
        super().enterEvent(event)

    def leaveEvent(self, event: QEvent) -> None:
        """
        Restore the widget's background color when not hovering.

        Parameters
        ----------
        event : QEvent
            The mouse leave event.

        Returns
        -------
        None
        """
        if self.INTERACT_ON_CLICK:
            self._css.QWidget.setValues(
                backgroundColor=SpyderPalette.COLOR_BACKGROUND_4
            )
            self.setStyleSheet(self._css.toString())

        super().leaveEvent(event)


class BaseTimerStatus(StatusBarWidget):
    """
    Base class for status bar widgets that update based on timers.
    """

    def __init__(self, parent: QWidget | None = None) -> None:
        """
        Base class for status bar widgets that update based on timers.

        Parameters
        ----------
        parent : QWidget | None, optional
            The parent widget of this one, or ``None`` (default).

        Returns
        -------
        None
        """
        self.timer = None  # Needs to come before parent call
        super().__init__(parent)
        self._interval = 2000

        # Widget setup
        fm = self.label_value.fontMetrics()
        self.label_value.setMinimumWidth(fm.width("000%"))

        # Setup
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_status)
        self.timer.start(self._interval)

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Handle the widget close event by stopping the timer.

        Parameters
        ----------
        event : QCloseEvent
            The widget close event.

        Returns
        -------
        None
        """
        self.timer.stop()
        super().closeEvent(event)

    def setVisible(self, value: bool) -> None:
        """
        Stop the timer if the widget is not visible.

        Parameters
        ----------
        value : bool
            ``True`` if the timer should be started, ``False`` to stop it.

        Returns
        -------
        None
        """
        if self.timer is not None:
            if value:
                self.timer.start(self._interval)
            else:
                self.timer.stop()
        super().setVisible(value)

    # ---- Public API
    # -------------------------------------------------------------------------
    def update_status(self) -> None:
        """
        Update the status label widget, if the widget is visible.

        Returns
        -------
        None
        """
        if self.isVisible():
            self.label_value.setText(self.get_value())

    def set_interval(self, interval: int) -> None:
        """
        Set the timer interval.

        Parameters
        ----------
        interval : int
            The timer interval, in integer milliseconds.

        Returns
        -------
        None
        """
        self._interval = interval
        if self.timer is not None:
            self.timer.setInterval(interval)

    def get_value(self) -> str:
        """
        Return the formatted text value shown in the widget.

        Returns
        -------
        str
            The formatted text value.

        Raises
        ------
        NotImplementedError
            Must be implemented by subclasses to work.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
String search and match utilities useful when filtering a list of texts.
"""

import re

NOT_FOUND_SCORE = -1
NO_SCORE = 0


def get_search_regex(query, ignore_case=True):
    """Returns a compiled regex pattern to search for query letters in order.

    Parameters
    ----------
    query : str
        String to search in another string (in order of character occurrence).
    ignore_case : True
        Optional value perform a case insensitive search (True by default).

    Returns
    -------
    pattern : SRE_Pattern

    Notes
    -----
    This function adds '.*' between the query characters and compiles the
    resulting regular expression.
    """
    regex_text = [char for char in query if char != ' ']
    regex_text = '.*'.join(regex_text)

    regex = u'({0})'.format(regex_text)

    if ignore_case:
        pattern = re.compile(regex, re.IGNORECASE)
    else:
        pattern = re.compile(regex)

    return pattern


def get_search_score(query, choice, ignore_case=True, apply_regex=True,
                     template='{}'):
    """Returns a tuple with the enriched text (if a template is provided) and
    a score for the match.

    Parameters
    ----------
    query : str
        String with letters to search in choice (in order of appearance).
    choice : str
        Sentence/words in which to search for the 'query' letters.
    ignore_case : bool, optional
        Optional value perform a case insensitive search (True by default).
    apply_regex : bool, optional
        Optional value (True by default) to perform a regex search. Useful
        when this function is called directly.
    template : str, optional
        Optional template string to surround letters found in choices. This is
        useful when using a rich text editor ('{}' by default).
        Examples: '<b>{}</b>', '<code>{}</code>', '<i>{}</i>'

    Returns
    -------
    results : tuple
        Tuples where the first item is the text (enriched if a template was
        used) and the second item is a search score.

    Notes
    -----
    The score is given according the following precedence (high to low):

    - Letters in one word and no spaces with exact match.
      Example: 'up' in 'up stroke'
    - Letters in one word and no spaces with partial match.
      Example: 'up' in 'upstream stroke'
    - Letters in one word but with skip letters.
      Example: 'cls' in 'close up'
    - Letters in two or more words
      Example: 'cls' in 'car lost'
    """
    original_choice = str(choice)
    result = (original_choice, NOT_FOUND_SCORE)

    # Handle empty string case
    if not query:
        return result

    query = str(query)
    choice = str(choice)

    if ignore_case:
        query = query.lower()
        choice = choice.lower()

    if apply_regex:
        pattern = get_search_regex(query, ignore_case=ignore_case)
        r = re.search(pattern, choice)
        if r is None:
            return result
    else:
        sep = u'-'  # Matches will be replaced by this character
        let = u'x'  # Nonmatches (except spaed) will be replaced by this
        score = 0

        exact_words = [query == str(word) for word in choice.split(' ')]
        partial_words = [query in word for word in choice.split(u' ')]

        if any(exact_words) or any(partial_words):
            pos_start = choice.find(query)
            pos_end = pos_start + len(query)
            score += pos_start
            text = choice.replace(query, sep*len(query), 1)

            enriched_text = original_choice[:pos_start] +\
                template.format(original_choice[pos_start:pos_end]) +\
                original_choice[pos_end:]

        if any(exact_words):
            # Check if the query words exists in a word with exact match
            score += 1
        elif any(partial_words):
            # Check if the query words exists in a word with partial match
            score += 100
        else:
            # Check letter by letter
            text = [l for l in original_choice]
            if ignore_case:
                temp_text = [l.lower() for l in original_choice]
            else:
                temp_text = text[:]

            # Give points to start of string
            score += temp_text.index(query[0])

            # Find the query letters and replace them by `sep`, also apply
            # template as needed for enricching the letters in the text
            enriched_text = text[:]
            for char in query:
                if char != u'' and char in temp_text:
                    index = temp_text.index(char)
                    enriched_text[index] = template.format(text[index])
                    text[index] = sep
                    temp_text = [u' ']*(index + 1) + temp_text[index+1:]

        enriched_text = u''.join(enriched_text)

        patterns_text = []
        for i, char in enumerate(text):
            if char != u' ' and char != sep:
                new_char = let
            else:
                new_char = char
            patterns_text.append(new_char)
        patterns_text = u''.join(patterns_text)
        for i in reversed(range(1, len(query) + 1)):
            score += (len(query) - patterns_text.count(sep*i))*100000

        temp = patterns_text.split(sep)
        while u'' in temp:
            temp.remove(u'')
        if not patterns_text.startswith(sep):
            temp = temp[1:]
        if not patterns_text.endswith(sep):
            temp = temp[:-1]

        for pat in temp:
            score += pat.count(u' ')*10000
            score += pat.count(let)*100

    return original_choice, enriched_text, score


def get_search_scores(query, choices, ignore_case=True, template='{}',
                      valid_only=False, sort=False):
    """Search for query inside choices and return a list of tuples.

    Returns a list of tuples of text with the enriched text (if a template is
    provided) and a score for the match. Lower scores imply a better match.

    Parameters
    ----------
    query : str
        String with letters to search in each choice (in order of appearance).
    choices : list of str
        List of sentences/words in which to search for the 'query' letters.
    ignore_case : bool, optional
        Optional value perform a case insensitive search (True by default).
    template : str, optional
        Optional template string to surround letters found in choices. This is
        useful when using a rich text editor ('{}' by default).
        Examples: '<b>{}</b>', '<code>{}</code>', '<i>{}</i>'

    Returns
    -------
    results : list of tuples
        List of tuples where the first item is the text (enriched if a
        template was used) and a search score. Lower scores means better match.
    """
    # First remove spaces from query
    query = query.replace(' ', '')
    pattern = get_search_regex(query, ignore_case)
    results = []

    for choice in choices:
        r = re.search(pattern, choice)
        if query and r:
            result = get_search_score(query, choice, ignore_case=ignore_case,
                                      apply_regex=False, template=template)
        else:
            if query:
                result = (choice, choice, NOT_FOUND_SCORE)
            else:
                result = (choice, choice, NO_SCORE)

        if valid_only:
            if result[-1] != NOT_FOUND_SCORE:
                results.append(result)
        else:
            results.append(result)

    if sort:
        results = sorted(results, key=lambda row: row[-1])

    return results


def test():
    template = '<b>{0}</b>'
    names = ['close pane', 'debug continue', 'debug exit', 'debug step into',
             'debug step over', 'debug step return', 'fullscreen mode',
             'layout preferences', 'lock unlock panes', 'maximize pane',
             'preferences', 'quit', 'restart', 'save current layout',
             'switch to breakpoints', 'switch to console', 'switch to editor',
             'switch to explorer', 'switch to find_in_files',
             'switch to historylog', 'switch to help',
             'switch to ipython_console', 'switch to onlinehelp',
             'switch to outline_explorer', 'switch to project_explorer',
             'switch to variable_explorer',
             'use next layout', 'use previous layout', 'clear line',
             'clear shell', 'inspect current object', 'blockcomment',
             'breakpoint', 'close all', 'code completion',
             'conditional breakpoint', 'configure', 'copy', 'copy line', 'cut',
             'debug', 'debug with winpdb', 'delete', 'delete line',
             'duplicate line', 'end of document', 'end of line',
             'file list management', 'find next', 'find previous', 'find text',
             'go to definition', 'go to line', 'go to next file',
             'go to previous file', 'inspect current object', 'kill next word',
             'kill previous word', 'kill to line end', 'kill to line start',
             'last edit location', 'move line down', 'move line up',
             'new file', 'next char', 'next cursor position', 'next line',
             'next word', 'open file', 'paste', 'previous char',
             'previous cursor position', 'previous line', 'previous word',
             'print', 're-run last script', 'redo', 'replace text',
             'rotate kill ring', 'run', 'run selection', 'save all', 'save as',
             'save file', 'select all', 'show/hide outline',
             'show/hide project explorer', 'start of document',
             'start of line', 'toggle comment', 'unblockcomment', 'undo',
             'yank', 'run profiler', 'run analysis']

    a = get_search_scores('lay', names, template=template, )
    b = get_search_scores('lay', names, template=template, valid_only=True,
                          sort=True)
    # Full results
    for r in a:
        print(r)  # spyder: test-skip

    # Ordered and filtered results
    print('\n')  # spyder: test-skip

    for r in b:
        print(r)  # spyder: test-skip

if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Find/Replace widget"""

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

# Standard library imports
import re
import sys

# Third party imports
from qtpy import PYSIDE2
from qtpy.QtCore import QEvent, QSize, Qt, QTimer, Signal, Slot
from qtpy.QtGui import QPixmap, QTextCursor
from qtpy.QtWidgets import (QAction, QGridLayout, QHBoxLayout, QLabel,
                            QLineEdit, QToolButton, QSizePolicy, QSpacerItem,
                            QWidget)

# Local imports
from spyder.api.shortcuts import SpyderShortcutsMixin
from spyder.api.translations import _
from spyder.utils.icon_manager import ima
from spyder.utils.misc import regexp_error_msg
from spyder.plugins.editor.utils.editor import TextHelper
from spyder.utils.qthelpers import create_toolbutton
from spyder.utils.sourcecode import get_eol_chars
from spyder.utils.stylesheet import AppStyle
from spyder.widgets.comboboxes import PatternComboBox


def is_position_sup(pos1, pos2):
    """Return True is pos1 > pos2"""
    return pos1 > pos2

def is_position_inf(pos1, pos2):
    """Return True is pos1 < pos2"""
    return pos1 < pos2


class SearchText(PatternComboBox):

    def __init__(self, parent):
        self.recommended_width = AppStyle.FindMinWidth
        super().__init__(parent, adjust_to_minimum=False)

    def sizeHint(self):
        """Recommended size."""
        return QSize(self.recommended_width, AppStyle.FindHeight)


class FindReplace(QWidget, SpyderShortcutsMixin):
    """Find widget"""

    # For shortcuts
    CONF_SECTION = 'find_replace'

    TOOLTIP = {
        'regexp_error': _("Regular expression error"),
        'no_matches': _("No matches")
    }

    visibility_changed = Signal(bool)
    return_shift_pressed = Signal()
    return_pressed = Signal()

    def __init__(self, parent, enable_replace=False):
        if not PYSIDE2:
            super().__init__(parent)
        else:
            QWidget.__init__(self, parent)
        SpyderShortcutsMixin.__init__(self)
        self.enable_replace = enable_replace
        self.editor = None
        self.is_code_editor = None

        glayout = QGridLayout()
        glayout.setContentsMargins(
            2 * AppStyle.MarginSize,
            AppStyle.MarginSize,
            2 * AppStyle.MarginSize,
            0
        )
        if sys.platform == "darwin":
            # Spacing is too big on Mac, which makes the widget look bad
            glayout.setSpacing(2 * AppStyle.MarginSize)
        self.setLayout(glayout)

        self.close_button = create_toolbutton(
            self,
            triggered=self.hide,
            icon=ima.icon('DialogCloseButton')
        )
        glayout.addWidget(self.close_button, 0, 0)

        # Icon size is the same for all buttons
        self.icon_size = self.close_button.iconSize()

        # Find layout
        self.search_text = SearchText(self)

        self.return_shift_pressed.connect(
            lambda:
            self.find(changed=False, forward=False, rehighlight=False,
                      multiline_replace_check = False)
        )

        self.return_pressed.connect(
            lambda:
            self.find(changed=False, forward=True, rehighlight=False,
                      multiline_replace_check = False)
        )

        self.search_text.lineEdit().textEdited.connect(
            self.text_has_been_edited)
        self.search_text.sig_resized.connect(self._resize_replace_text)

        self.number_matches_text = QLabel(self)
        self.search_text.lineEdit().clear_action.triggered.connect(
            self.clear_matches
        )
        self.hide_number_matches_text = False
        self.number_matches_pixmap = (
            ima.icon('number_matches').pixmap(self.icon_size)
        )
        self.matches_string = ""

        self.no_matches_icon = ima.icon('no_matches')
        self.error_icon = ima.icon('error')
        self.messages_action = QAction(self)
        self.messages_action.setVisible(False)
        self.search_text.lineEdit().addAction(
            self.messages_action, QLineEdit.TrailingPosition)

        # Button corresponding to the messages_action above
        self.messages_button = (
            self.search_text.lineEdit().findChildren(QToolButton)[1]
        )

        self.replace_on = False
        self.replace_text_button = create_toolbutton(
            self,
            toggled=self.change_replace_state,
            icon=ima.icon('replace'),
            tip=_("Replace text")
        )
        if not self.enable_replace:
            self.replace_text_button.hide()

        self.previous_button = create_toolbutton(
            self,
            triggered=self.find_previous,
            icon=ima.icon('findprevious'),
            tip=_("Find previous")
        )

        self.next_button = create_toolbutton(
            self,
            triggered=self.find_next,
            icon=ima.icon('findnext'),
            tip=_("Find next")
        )
        self.next_button.clicked.connect(self.update_search_combo)
        self.previous_button.clicked.connect(self.update_search_combo)

        self.re_button = create_toolbutton(
            self, icon=ima.icon('regex'),
            tip=_("Use regular expressions")
        )
        self.re_button.setCheckable(True)
        self.re_button.toggled.connect(lambda state: self.find())

        self.case_button = create_toolbutton(
            self,
            icon=ima.icon("format_letter_case"),
            tip=_("Enable case sensitive searches")
        )
        self.case_button.setCheckable(True)
        self.case_button.toggled.connect(lambda state: self.find())

        self.words_button = create_toolbutton(
            self,
            icon=ima.icon("format_letter_matches"),
            tip=_("Only search for whole words")
        )
        self.words_button.setCheckable(True)
        self.words_button.toggled.connect(lambda state: self.find())

        self.widgets = [
            self.close_button,
            self.search_text,
            self.previous_button,
            self.next_button,
            self.re_button,
            self.case_button,
            self.words_button,
            self.replace_text_button,
            self.number_matches_text,
        ]

        # Search layout
        search_layout = QHBoxLayout()
        for widget in self.widgets[1:-1]:
            search_layout.addWidget(widget)

        search_layout.addSpacerItem(QSpacerItem(10, 0))
        search_layout.addWidget(self.number_matches_text)
        search_layout.addSpacerItem(
            QSpacerItem(6, 0, QSizePolicy.Expanding)
        )

        glayout.addLayout(search_layout, 0, 1)

        # Replace layout
        self.replace_text = PatternComboBox(
            self,
            adjust_to_minimum=False
        )
        self.replace_text.valid.connect(
            lambda _: self.replace_find(focus_replace_text=True))
        self.replace_text.lineEdit().setPlaceholderText(_("Replace"))

        self.replace_button = create_toolbutton(
            self,
            tip=_('Replace next occurrence'),
            icon=ima.icon('replace_next'),
            triggered=self.replace_find,
        )

        self.replace_sel_button = create_toolbutton(
            self,
            tip=_('Replace occurrences in selection'),
            icon=ima.icon('replace_selection'),
            triggered=self.replace_find_selection,
        )
        self.replace_sel_button.clicked.connect(self.update_replace_combo)
        self.replace_sel_button.clicked.connect(self.update_search_combo)

        self.replace_all_button = create_toolbutton(
            self,
            tip=_('Replace all occurrences'),
            icon=ima.icon('replace_all'),
            triggered=self.replace_find_all,
        )
        self.replace_all_button.clicked.connect(self.update_replace_combo)
        self.replace_all_button.clicked.connect(self.update_search_combo)

        replace_layout = QHBoxLayout()
        widgets = [
            self.replace_text,
            self.replace_button,
            self.replace_sel_button,
            self.replace_all_button
        ]
        for widget in widgets:
            replace_layout.addWidget(widget)
        replace_layout.addStretch(1)
        glayout.addLayout(replace_layout, 1, 1)
        self.widgets.extend(widgets)
        self.replace_widgets = widgets
        self.hide_replace()

        # Additional adjustments
        self.search_text.setTabOrder(self.search_text, self.replace_text)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.register_shortcuts(parent)

        # To highlight found results in the editor
        self.highlight_timer = QTimer(self)
        self.highlight_timer.setSingleShot(True)
        self.highlight_timer.setInterval(300)
        self.highlight_timer.timeout.connect(self.highlight_matches)

        # Install event filter for search_text
        self.search_text.installEventFilter(self)

        # To avoid painting number_matches_text on every resize event
        self.show_matches_timer = QTimer(self)
        self.show_matches_timer.setSingleShot(True)
        self.show_matches_timer.setInterval(25)
        self.show_matches_timer.timeout.connect(self.show_matches)

    def eventFilter(self, widget, event):
        """
        Event filter for search_text widget.

        Notes
        -----
        * Emit signals when Enter and Shift+Enter are pressed. These signals
          are used for search forward and backward.
        * Add crude hack to get tab working between the find/replace boxes.
        * Reduce space between the messages_button and the clear one.
        """

        # Type check: Prevent error in PySide where 'event' may be of type
        # QtGui.QPainter (for whatever reason).
        if not isinstance(event, QEvent):
            return True

        if event.type() == QEvent.KeyPress:
            key = event.key()
            shift = event.modifiers() & Qt.ShiftModifier

            if key in (Qt.Key_Return, Qt.Key_Enter):
                if shift:
                    self.return_shift_pressed.emit()
                else:
                    self.return_pressed.emit()

            if key == Qt.Key_Tab:
                if self.search_text.hasFocus():
                    self.replace_text.set_current_text(
                        self.search_text.currentText())
                self.focusNextChild()

        if event.type() == QEvent.Paint:
            self.messages_button.move(
                self.search_text.lineEdit().width() - 42,
                self.messages_button.y()
            )

        return super().eventFilter(widget, event)

    def register_shortcuts(self, parent):
        """Register shortcuts for this widget."""
        shortcuts = (
            ('find next', self.find_next, parent),
            ("find previous", self.find_previous, parent),
            ('find text', self.show, parent),
            ('replace text', self.show_replace, parent),
            ('replace all', self.replace_find_all, parent),
            ('hide find and replace', self.hide, self),
        )

        for name, callback, widget in shortcuts:
            self.register_shortcut_for_widget(
                name=name, triggered=callback, widget=widget
            )

    def update_search_combo(self):
        self.search_text.lineEdit().returnPressed.emit()

    def update_replace_combo(self):
        self.replace_text.lineEdit().returnPressed.emit()

    def show(self, hide_replace=True):
        """Overrides Qt Method"""
        QWidget.show(self)

        self._width_adjustments()
        self.visibility_changed.emit(True)
        self.change_number_matches()

        if self.editor is not None:
            if hide_replace:
                if self.replace_widgets[0].isVisible():
                    self.hide_replace()
            else:
                self.replace_text_button.setChecked(True)

            # When selecting several lines, and replace box is activated the
            # text won't be replaced for the selection
            text = self.editor.get_selected_text()
            if hide_replace or len(text.splitlines()) <= 1:
                highlighted = True
                # If no text is highlighted for search, use whatever word is
                # under the cursor
                if not text:
                    highlighted = False
                    try:
                        cursor = self.editor.textCursor()
                        cursor.select(QTextCursor.WordUnderCursor)
                        text = str(cursor.selectedText())
                    except AttributeError:
                        # We can't do this for all widgets, e.g. WebView's
                        pass

                # Now that text value is sorted out, use it for the search
                if text and not self.search_text.currentText() or highlighted:
                    self.search_text.setEditText(text)
                    self.search_text.lineEdit().selectAll()
                    self.refresh()
                else:
                    self.search_text.lineEdit().selectAll()
            self.search_text.setFocus()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._width_adjustments()

    @Slot()
    def replace_widget(self, replace_on):
        """Show and hide replace widget"""
        if replace_on:
            self.show_replace()
        else:
            self.hide_replace()

    def change_replace_state(self):
        """Handle the change of the replace state widget."""
        self.replace_on = not self.replace_on
        self.replace_text_button.setChecked(self.replace_on)
        self.replace_widget(self.replace_on)

    def hide(self):
        """Overrides Qt Method"""
        for widget in self.replace_widgets:
            widget.hide()
        QWidget.hide(self)
        self.replace_text_button.setChecked(False)
        self.visibility_changed.emit(False)
        if self.editor is not None:
            self.editor.setFocus()
            self.clear_matches()

    def show_replace(self):
        """Show replace widgets"""
        if self.enable_replace:
            self.show(hide_replace=False)
            for widget in self.replace_widgets:
                widget.show()

    def hide_replace(self):
        """Hide replace widgets"""
        for widget in self.replace_widgets:
            widget.hide()
            self.replace_text_button.setChecked(False)

    def refresh(self):
        """Refresh widget"""
        if self.isHidden():
            if self.editor is not None:
                self.clear_matches()
            return
        state = self.editor is not None
        for widget in self.widgets:
            widget.setEnabled(state)
        if state:
            self.find()

    def set_editor(self, editor, refresh=True):
        """Set associated editor."""
        # Note: This is necessary to test widgets/editor.py in Qt builds that
        # don't have web widgets
        try:
            from qtpy.QtWebEngineWidgets import QWebEngineView
        except ImportError:
            QWebEngineView = type(None)
        from spyder.plugins.editor.widgets.codeeditor import CodeEditor

        self.words_button.setVisible(not isinstance(editor, QWebEngineView))
        self.re_button.setVisible(not isinstance(editor, QWebEngineView))
        self.is_code_editor = isinstance(editor, CodeEditor)

        # Disconnect previous connection to highlight matches
        if self.editor is not None and self.is_code_editor:
            self.editor.textChanged.disconnect(self.update_matches)

        # Set current editor
        self.editor = editor

        # Keep number of matches updated if editor text has changed
        if self.is_code_editor:
            self.editor.textChanged.connect(self.update_matches)

        if refresh:
            self.refresh()
        if self.isHidden() and editor is not None:
            self.clear_matches()

    @Slot()
    def find_next(self, set_focus=True):
        """Find next occurrence"""
        state = self.find(changed=False, forward=True, rehighlight=False,
                          multiline_replace_check=False)
        if set_focus:
            self.editor.setFocus()
        self.search_text.add_current_text()
        return state

    @Slot()
    def find_previous(self, set_focus=True):
        """Find previous occurrence"""
        state = self.find(changed=False, forward=False, rehighlight=False,
                          multiline_replace_check=False)
        if set_focus:
            self.editor.setFocus()
        return state

    def text_has_been_edited(self, text):
        """
        Find text has been edited (this slot won't be triggered when setting
        the search pattern combo box text programmatically).
        """
        self.find(changed=True, forward=True, start_highlight_timer=True)

    def highlight_matches(self):
        """Highlight found results"""
        if self.is_code_editor:
            text = self.search_text.currentText()
            case = self.case_button.isChecked()
            word = self.words_button.isChecked()
            regexp = self.re_button.isChecked()
            self.editor.highlight_found_results(
                text, word=word, regexp=regexp, case=case)

    def clear_matches(self):
        """Clear all highlighted matches"""
        self.matches_string = ""
        self.messages_action.setVisible(False)
        self.number_matches_text.hide()
        if self.is_code_editor:
            self.editor.clear_found_results()

    def find(self, changed=True, forward=True, rehighlight=True,
             start_highlight_timer=False, multiline_replace_check=True):
        """Call the find function"""
        # When several lines are selected in the editor and replace box is
        # activated, dynamic search is deactivated to prevent changing the
        # selection. Otherwise we show matching items.
        if multiline_replace_check and self.replace_widgets[0].isVisible():
            sel_text = self.editor.get_selected_text()
            if len(str(sel_text).splitlines()) > 1:
                return None
        text = self.search_text.currentText()
        if len(text) == 0:
            if not self.is_code_editor:
                # Clears the selection for WebEngine
                self.editor.find_text('')
            self.change_number_matches()
            self.clear_matches()
            return None
        else:
            case = self.case_button.isChecked()
            word = self.words_button.isChecked()
            regexp = self.re_button.isChecked()
            found = self.editor.find_text(text, changed, forward, case=case,
                                          word=word, regexp=regexp)

            error_msg = False
            if not found and regexp:
                error_msg = regexp_error_msg(text)
                if error_msg:
                    self.show_error(error_msg)

            # No need to continue after this point if we detected an error in
            # the passed regexp.
            if error_msg:
                return

            if self.is_code_editor and found:
                cursor = QTextCursor(self.editor.textCursor())
                TextHelper(self.editor).unfold_if_colapsed(cursor)

                if rehighlight or not self.editor.found_results:
                    self.highlight_timer.stop()
                    if start_highlight_timer:
                        self.highlight_timer.start()
                    else:
                        self.highlight_matches()
            else:
                self.clear_matches()

            number_matches = self.editor.get_number_matches(text, case=case,
                                                            regexp=regexp,
                                                            word=word)
            if hasattr(self.editor, 'get_match_number'):
                match_number = self.editor.get_match_number(text, case=case,
                                                            regexp=regexp,
                                                            word=word)
            else:
                match_number = 0
            self.change_number_matches(current_match=match_number,
                                       total_matches=number_matches)
            return found

    @Slot()
    def replace_find(self, focus_replace_text=False):
        """Replace and find."""
        if self.editor is None:
            return
        replace_text = str(self.replace_text.currentText())
        search_text = str(self.search_text.currentText())
        re_pattern = None
        case = self.case_button.isChecked()
        re_flags = re.MULTILINE if case else re.IGNORECASE | re.MULTILINE

        # Check regexp before proceeding
        if self.re_button.isChecked():
            try:
                re_pattern = re.compile(search_text, flags=re_flags)
                # Check if replace_text can be substituted in re_pattern
                # Fixes spyder-ide/spyder#7177.
                re_pattern.sub(replace_text, '')
            except re.error:
                # Do nothing with an invalid regexp
                return

        # First found
        seltxt = str(self.editor.get_selected_text())
        cmptxt1 = search_text if case else search_text.lower()
        cmptxt2 = seltxt if case else seltxt.lower()
        do_replace = True
        if re_pattern is None:
            has_selected = self.editor.has_selected_text()
            if not has_selected or cmptxt1 != cmptxt2:
                if not self.find(changed=False, forward=True,
                                 rehighlight=False):
                    do_replace = False
        else:
            if len(re_pattern.findall(cmptxt2)) <= 0:
                if not self.find(changed=False, forward=True,
                                 rehighlight=False):
                    do_replace = False
        cursor = None
        if do_replace:
            cursor = self.editor.textCursor()
            cursor.beginEditBlock()

            if re_pattern is None:
                cursor.removeSelectedText()
                cursor.insertText(replace_text)
            else:
                seltxt = str(cursor.selectedText())

                # Note: If the selection obtained from an editor spans a line
                # break, the text will contain a Unicode U+2029 paragraph
                # separator character instead of a newline \n character.
                # See: spyder-ide/spyder#2675
                eol_char = get_eol_chars(self.editor.toPlainText())
                seltxt = seltxt.replace(u'\u2029', eol_char)

                cursor.removeSelectedText()
                cursor.insertText(re_pattern.sub(replace_text, seltxt))

            if self.find_next(set_focus=False):
                found_cursor = self.editor.textCursor()
                cursor.setPosition(found_cursor.selectionStart(),
                                   QTextCursor.MoveAnchor)
                cursor.setPosition(found_cursor.selectionEnd(),
                                   QTextCursor.KeepAnchor)

        if cursor is not None:
            cursor.endEditBlock()

        if focus_replace_text:
            self.replace_text.setFocus()
        else:
            self.editor.setFocus()

        if getattr(self.editor, 'document_did_change', False):
            self.editor.document_did_change()

    @Slot()
    def replace_find_all(self):
        """Replace and find all matching occurrences"""
        if self.editor is None:
            return

        replace_text = str(self.replace_text.currentText())
        search_text = str(self.search_text.currentText())
        case = self.case_button.isChecked()
        word = self.words_button.isChecked()
        re_flags = re.MULTILINE if case else re.IGNORECASE | re.MULTILINE

        if self.re_button.isChecked():
            pattern = search_text
        else:
            pattern = re.escape(search_text)
            # re.sub processes backslashes so they must be escaped
            # See spyder-ide/spyder#21007.
            replace_text = replace_text.replace('\\', r'\\')

        # Match whole words only
        if word:
            pattern = r'\b{pattern}\b'.format(pattern=pattern)

        # Check regexp before proceeding
        re_pattern = None
        try:
            re_pattern = re.compile(pattern, flags=re_flags)

            # Check if replace_text can be substituted in re_pattern
            # Fixes spyder-ide/spyder#7177.
            re_pattern.sub(replace_text, '')
        except re.error:
            # Do nothing with an invalid regexp
            return

        cursor = self.editor._select_text("sof", "eof")
        text = self.editor.toPlainText()
        cursor.beginEditBlock()
        cursor.removeSelectedText()
        cursor.insertText(re_pattern.sub(replace_text, text))
        cursor.endEditBlock()

        self.editor.setFocus()

    @Slot()
    def replace_find_selection(self, focus_replace_text=False):
        """Replace and find in the current selection"""
        if self.editor is not None:
            replace_text = str(self.replace_text.currentText())
            search_text = str(self.search_text.currentText())
            case = self.case_button.isChecked()
            word = self.words_button.isChecked()
            re_flags = re.MULTILINE if case else re.IGNORECASE | re.MULTILINE

            re_pattern = None
            if self.re_button.isChecked():
                pattern = search_text
            else:
                pattern = re.escape(search_text)
                # re.sub processes backslashes so they must be escaped
                replace_text = replace_text.replace('\\', r'\\')
            if word:  # match whole words only
                pattern = r'\b{pattern}\b'.format(pattern=pattern)

            # Check regexp before proceeding
            try:
                re_pattern = re.compile(pattern, flags=re_flags)
                # Check if replace_text can be substituted in re_pattern
                # Fixes spyder-ide/spyder#7177.
                re_pattern.sub(replace_text, '')
            except re.error:
                # Do nothing with an invalid regexp
                return

            selected_text = str(self.editor.get_selected_text())
            replacement = re_pattern.sub(replace_text, selected_text)
            if replacement != selected_text:
                cursor = self.editor.textCursor()
                start_pos = cursor.selectionStart()
                cursor.beginEditBlock()
                cursor.removeSelectedText()
                cursor.insertText(replacement)
                # Restore selection
                self.editor.set_cursor_position(start_pos)
                for c in range(len(replacement)):
                    self.editor.extend_selection_to_next('character', 'right')
                cursor.endEditBlock()

            if focus_replace_text:
                self.replace_text.setFocus()
            else:
                self.editor.setFocus()

            if getattr(self.editor, 'document_did_change', False):
                self.editor.document_did_change()

    def change_number_matches(self, current_match=0, total_matches=0):
        """Change number of match and total matches."""
        if current_match and total_matches:
            self.matches_string = "{} {} {}".format(current_match, _("of"),
                                                    total_matches)
            self.show_matches()
        elif total_matches:
            self.matches_string = "{} {}".format(total_matches, _("matches"))
            self.show_matches()
        else:
            self.number_matches_text.hide()
            if self.search_text.currentText():
                self.show_no_matches()

    def update_matches(self):
        """Update total number of matches if text has changed in the editor."""
        if self.isVisible():
            number_matches = self.editor.get_number_matches(
                self.search_text.lineEdit().text(),
                case=self.case_button.isChecked(),
                regexp=self.re_button.isChecked(),
                word=self.words_button.isChecked()
            )
            self.change_number_matches(total_matches=number_matches)

    def show_no_matches(self):
        """Show a no matches message with an icon."""
        self._show_icon_message('no_matches')

    def show_matches(self):
        """Show the number of matches found in the document."""
        if not self.matches_string:
            return

        self.number_matches_text.show()
        self.messages_action.setVisible(False)

        if self.hide_number_matches_text:
            self.number_matches_text.setPixmap(self.number_matches_pixmap)
            self.number_matches_text.setToolTip(self.matches_string)
        else:
            self.number_matches_text.setPixmap(QPixmap())
            self.number_matches_text.setText(self.matches_string)

    def show_error(self, error_msg):
        """Show a regexp error message with an icon."""
        self._show_icon_message('error', extra_info=error_msg)

    def _show_icon_message(self, kind, extra_info=None):
        """
        Show a message to users with an icon when no matches can be found or
        there's an error in the passed regexp.

        Parameters
        ----------
        kind: str
            The kind of message. It can be 'no_matches' or 'error'.
        extra_info:
            Extra info to add to the icon's tooltip.
        """
        if kind == 'no_matches':
            tooltip = self.TOOLTIP['no_matches']
            icon = self.no_matches_icon
        else:
            tooltip = self.TOOLTIP['regexp_error']
            icon = self.error_icon

        if extra_info:
            tooltip = tooltip + ': ' + extra_info

        self.messages_action.setIcon(icon)
        self.messages_action.setToolTip(tooltip)
        self.messages_action.setVisible(True)

    def _width_adjustments(self):
        """Several adjustments according to the widget's total width."""
        # The widgets list includes search_text and number_matches_text. That's
        # why we substract a 2 below.
        buttons_width = self.icon_size.width() * (len(self.widgets) - 2)

        total_width = self.size().width()
        matches_width = self.number_matches_text.size().width()
        minimal_width = (
            self.search_text.recommended_width + buttons_width + matches_width
        )

        if total_width < minimal_width:
            self.search_text.setMinimumWidth(30)
            self.hide_number_matches_text = True
        else:
            self.search_text.setMinimumWidth(int(total_width / 2))
            self.hide_number_matches_text = False

        # We don't call show_matches directly here to avoid flickering when the
        # user hits the widget's minimal width, which changes from text to an
        # icon (or vice versa) for number_matches_text.
        self.show_matches_timer.start()

    def _resize_replace_text(self, size, old_size):
        """
        Resize replace_text combobox to match the width of the search one.
        """
        self.replace_text.setMinimumWidth(size.width())
        self.replace_text.setMaximumWidth(size.width())



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Base classes for constructing and accessing top-level Spyder plugin objects.

Here, "plugins" are Qt objects that can make changes to Spyder's main window
and call/connect to other plugins directly.

There are two types of plugins available:

* :class:`SpyderPluginV2` does not create a new pane in Spyder's main window.

   .. note::

       :class:`SpyderPluginV2` will be aliased to :class:`!SpyderPlugin`
       and deprecated in Spyder 6.2, and removed in Spyder 7.0.

* :class:`SpyderDockablePlugin` creates a new pane in Spyder's main window.
"""

from .enum import DockablePlugins, OptionalPlugins, Plugins  # noqa
from .new_api import SpyderDockablePlugin, SpyderPluginV2  # noqa


__all__ = [
    "Plugins",
    "DockablePlugins",
    "OptionalPlugins",
    "SpyderPluginV2",
    "SpyderDockablePlugin",
]



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""IPython Console config page."""

# Standard library imports
import sys

# Third party imports
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QVBoxLayout,
)

# Local imports
from spyder.api.translations import _
from spyder.api.preferences import PluginConfigPage


class IPythonConsoleConfigPage(PluginConfigPage):

    def __init__(self, plugin, parent):
        super().__init__(plugin, parent)

        self.buffer_spin = None
        self.apply_callback = self.warn_if_large_buffer

    def setup_page(self):
        newcb = self.create_checkbox

        # Display group
        display_group = QGroupBox(_("Display"))
        banner_box = newcb(
            _("Show welcome message"),
            'show_banner',
            tip=_("Print the startup message when opening a new console"),
        )
        calltips_box = newcb(
            _("Show calltips"),
            'show_calltips',
            tip=_("Show a summary help popup when typing an open parenthesis "
                  "after a callable function"),
        )
        show_time_box = newcb(
            _("Show elapsed time"),
            'show_elapsed_time',
            tip=_("Display the time since the current console was started "
                  "in the tab bar"),
        )

        display_layout = QVBoxLayout()
        display_layout.addWidget(banner_box)
        display_layout.addWidget(calltips_box)
        display_layout.addWidget(show_time_box)
        display_group.setLayout(display_layout)

        # Confirmation group
        confirmations_group = QGroupBox(_("Confirmation"))
        ask_box = newcb(
            _("Ask for confirmation before closing"),
            'ask_before_closing',
        )
        reset_namespace_box = newcb(
            _("Ask for confirmation before removing all variables"),
            'show_reset_namespace_warning',
        )
        ask_restart_box = newcb(
            _("Ask for confirmation before restarting"),
            'ask_before_restart',
        )

        confirmations_layout = QVBoxLayout()
        confirmations_layout.addWidget(ask_box)
        confirmations_layout.addWidget(ask_restart_box)
        confirmations_layout.addWidget(reset_namespace_box)
        confirmations_group.setLayout(confirmations_layout)

        # Completion group
        comp_group = QGroupBox(_("Completion"))
        completers = [(_("Graphical"), 0), (_("Terminal"), 1), (_("Plain"), 2)]
        comp_box = self.create_combobox(
            _("Display:"),
            completers,
            'completion_type',
            tip=_(
                "Graphical shows a list of completion matches in a GUI.\n"
                "Plain displays matches in the Console output, like Bash.\n"
                "Terminal is Plain plus Tab selecting matches, like Zsh.\n"
            ),
        )
        jedi_box = newcb(
            _("Use Jedi completion"),
            "jedi_completer",
            tip=_(
                "Enable Jedi-based tab-completion in the IPython console.\n"
                "Similar to the greedy completer, but without evaluating "
                "the code and allows completion of dictionary keys, "
                "nested lists and similar.\n"
                "Warning: Can slow down the Console when working with "
                "large dataframes."
            ),
        )
        greedy_box = newcb(
            _("Use greedy completion"),
            "greedy_completer",
            tip=_(
                "Enable <tt>Tab</tt> completion on elements of lists, "
                "results of function calls and similar "
                "<i>without</i> assigning them to a variable, "
                "like <tt>li[0].&lt;Tab&gt;</tt> or "
                "<tt>ins.meth().&lt;Tab&gt;</tt><br>"
                "<b>Warning</b>: This can be unsafe because your code "
                "is actually executed when you press <tt>Tab</tt>."
            ),
        )

        comp_layout = QVBoxLayout()
        comp_layout.addWidget(comp_box)
        comp_layout.addWidget(jedi_box)
        comp_layout.addWidget(greedy_box)
        comp_group.setLayout(comp_layout)

        # Output group
        output_group = QGroupBox(_("Output"))
        self.buffer_spin = self.create_spinbox(
            _("Buffer:"),
            _(" lines"),
            'buffer_size',
            min_=100,
            # >10k can make Spyder slow, see spyder-ide/spyder#19091
            max_=50_000,
            step=100,
            tip=_(
                "The maximum number of output lines "
                "retained in each console at a time.\n"
                "Warning; Buffer sizes greater than 10000 lines can slow "
                "down Spyder."
            ),
        )
        sympy_box = newcb(
            _("Render SymPy symbolic math"),
            "symbolic_math",
            tip=_(
                "Pretty-print the outputs of SymPy symbolic computations\n"
                "(requires SymPy installed in the console environment).\n"
                "Refer to SymPy's documentation for details on using it."
            ),
        )

        output_layout = QVBoxLayout()
        output_layout.addWidget(self.buffer_spin)
        output_layout.addWidget(sympy_box)
        output_group.setLayout(output_layout)

        # --- Plotting ---
        # Matplotlib group
        matplotlib_group = QGroupBox(_("Matplotlib support"))
        matplotlib_box = newcb(_("Activate support"), 'pylab')
        autoload_matplotlib_box = newcb(
            _("Automatically import NumPy and Matplotlib modules"),
            'pylab/autoload',
            tip=_(
                "This is a convenience to use NumPy and Matplotlib\n"
                "in the console without explicitly importing the modules."
            )
        )
        autoload_matplotlib_box.setEnabled(self.get_option('pylab'))
        matplotlib_box.checkbox.toggled.connect(
            autoload_matplotlib_box.setEnabled
        )

        matplotlib_layout = QVBoxLayout()
        matplotlib_layout.addWidget(matplotlib_box)
        matplotlib_layout.addWidget(autoload_matplotlib_box)
        matplotlib_group.setLayout(matplotlib_layout)

        # Graphics backend group
        inline = _("Inline")
        automatic = _("Automatic")
        backend_group = QGroupBox(_("Graphics backend"))
        backend_label = QLabel(_("Choose how figures are displayed"))

        backends = [
            (inline, 'inline'),
            (automatic, 'auto'),
            ("Qt", 'qt'),
            ("Tk", 'tk'),
        ]

        if sys.platform == 'darwin':
            backends.append(("macOS", 'osx'))
        backends = tuple(backends)

        backend_box = self.create_combobox(
            _("Backend:"),
            backends,
            'pylab/backend',
            default='inline',
            tip=_(
                "If unsure, select {inline} to show figures in the Plots pane"
                "\nor {auto} to interact with them (zoom and pan) "
                "in a new window."
            ).format(inline=inline, auto=automatic),
        )

        backend_layout = QVBoxLayout()
        backend_layout.addWidget(backend_label)
        backend_layout.addWidget(backend_box)
        backend_group.setLayout(backend_layout)
        backend_group.setEnabled(self.get_option('pylab'))
        matplotlib_box.checkbox.toggled.connect(backend_group.setEnabled)

        # Inline backend group
        inline_group = QGroupBox(_("Inline backend"))
        inline_label = QLabel(_("Settings for figures in the Plots pane"))
        inline_label.setWordWrap(True)
        formats = (("PNG", 'png'), ("SVG", 'svg'))
        format_box = self.create_combobox(
            _("Format:") + "   ",
            formats,
            'pylab/inline/figure_format',
            default='png',
            tip=_(
                "PNG is more widely supported, "
                "while SVG is resolution-independent and easier to edit "
                "but complex plots may not be displayed correctly."
            ),
        )
        resolution_spin = self.create_spinbox(
            _("Resolution:") + "  ",
            " " + _("DPI"),
            'pylab/inline/resolution',
            min_=50,
            max_=999,
            step=0.1,
            tip=_("Only used when the format is PNG. Default is 144."),
        )
        width_spin = self.create_spinbox(
            _("Width:") + "  ",
            " " + _("inches"),
            'pylab/inline/width',
            min_=2,
            max_=20,
            step=1,
            tip=_("Default is 6"),
        )
        height_spin = self.create_spinbox(
            _("Height:") + "  ",
            " " + _("inches"),
            'pylab/inline/height',
            min_=1,
            max_=20,
            step=1,
            tip=_("Default is 4"),
        )
        fontsize_spin = self.create_spinbox(
            _("Font size:") + "  ",
            " " + _("points"),
            'pylab/inline/fontsize',
            min_=5,
            max_=48,
            step=1.0,
            tip=_("Default is 10"),
        )
        bottom_spin = self.create_spinbox(
            _("Bottom edge:") + "  ",
            " " + _("of figure height"),
            'pylab/inline/bottom',
            min_=0,
            max_=0.3,
            step=0.01,
            tip=_("The position of the bottom edge of the subplots,\n"
                  "as a fraction of the figure height (default is 0.11).")
        )
        bottom_spin.spinbox.setDecimals(2)
        bbox_inches_box = newcb(
            _("Use a tight layout for inline plots"),
            'pylab/inline/bbox_inches',
            tip=_("Sets 'bbox_inches' to 'tight' for inline plots.\n"
                  "When enabled, figures displayed in the Plots pane\n"
                  "may look different from those output with 'savefig'."))

        inline_v_layout = QVBoxLayout()
        inline_v_layout.addWidget(inline_label)
        inline_layout = QGridLayout()
        inline_layout.addWidget(format_box.label, 1, 0)
        inline_layout.addWidget(format_box.combobox, 1, 1)
        inline_layout.addWidget(format_box.help_label, 1, 3)

        spinboxes = [resolution_spin, width_spin, height_spin,
                     fontsize_spin, bottom_spin]
        for counter, spinbox in enumerate(spinboxes):
            inline_layout.addWidget(spinbox.plabel, counter + 2, 0)
            inline_layout.addWidget(spinbox.spinbox, counter + 2, 1)
            inline_layout.addWidget(spinbox.slabel, counter + 2, 2)
            inline_layout.addWidget(spinbox.help_label, counter + 2, 3)

        inline_layout.addWidget(bbox_inches_box, len(spinboxes) + 2, 0, 1, 4)

        inline_h_layout = QHBoxLayout()
        inline_h_layout.addLayout(inline_layout)
        inline_h_layout.addStretch(1)
        inline_v_layout.addLayout(inline_h_layout)
        inline_group.setLayout(inline_v_layout)
        inline_group.setEnabled(self.get_option('pylab'))
        matplotlib_box.checkbox.toggled.connect(inline_group.setEnabled)

        # --- Startup ---
        # Run lines group
        run_lines_group = QGroupBox(_("Run code"))
        run_lines_label = QLabel(_(
            "Enter a code snippet to run when a new console is started.\n"
            "Separate multiple lines by semicolons, for example:<br>"
            "<tt>import os; import sys</tt>"
        ))
        run_lines_label.setWordWrap(True)
        run_lines_edit = self.create_lineedit(_("Lines:"), 'startup/run_lines',
                                              '', alignment=Qt.Horizontal)

        run_lines_layout = QVBoxLayout()
        run_lines_layout.addWidget(run_lines_label)
        run_lines_layout.addWidget(run_lines_edit)
        run_lines_group.setLayout(run_lines_layout)

        # Run file group
        run_file_group = QGroupBox(_("Run a file"))
        run_file_label = QLabel(_(
            "Specify a Python file to execute at startup, similar to "
            "<tt>PYTHONSTARTUP</tt>"
        ))
        run_file_label.setWordWrap(True)
        file_radio = newcb(
            _("Execute the following file:"), 'startup/use_run_file', False
        )
        run_file_browser = self.create_browsefile('', 'startup/run_file', '')
        run_file_browser.setEnabled(False)
        file_radio.checkbox.toggled.connect(run_file_browser.setEnabled)

        run_file_layout = QVBoxLayout()
        run_file_layout.addWidget(run_file_label)
        run_file_layout.addWidget(file_radio)
        run_file_layout.addWidget(run_file_browser)
        run_file_group.setLayout(run_file_layout)

        # ---- Advanced settings ----
        # Autocall group
        autocall_group = QGroupBox(_("Autocall"))
        autocall_label = QLabel(_(
            "Implicitly insert parethesis after any callable object, "
            "treating anything following it as arguments.<br>"
            "For example, typing <tt>print 'Number:', 42</tt> will execute "
            "<tt>print('Number:', 42)</tt>."
        ))
        autocall_label.setWordWrap(True)

        smart = _('Smart')
        full = _('Full')
        autocall_opts = ((_('Off'), 0), (smart, 1), (full, 2))
        autocall_box = self.create_combobox(
            _("Autocall:  "),
            autocall_opts,
            'autocall',
            default=0,
            tip=_(
                "In {smart} mode, Autocall is not applied if there are no "
                "arguments after the callable.\n"
                "In {full} mode, callable objects are called even if no "
                "arguments are present."
            ).format(smart=smart, full=full),
        )

        autocall_layout = QVBoxLayout()
        autocall_layout.addWidget(autocall_label)
        autocall_layout.addWidget(autocall_box)
        autocall_group.setLayout(autocall_layout)

        # Autoreload group
        autoreload_group = QGroupBox(_("Autoreload"))
        autoreload_label = QLabel(_(
            "Reload imported modules automatically before running code. "
            "This is a different mechanism than the User Module Reloader "
            "and can be slow on Windows due to limitations of its file system."
        ))
        autoreload_label.setWordWrap(True)

        autoreload_box = newcb(
            _("Use autoreload"),
            "autoreload",
            tip=_(
                "Enables the autoreload magic. Refer to its documentation to "
                "learn how to use it."
            )
        )

        autoreload_layout = QVBoxLayout()
        autoreload_layout.addWidget(autoreload_label)
        autoreload_layout.addWidget(autoreload_box)
        autoreload_group.setLayout(autoreload_layout)

        # Prompts group
        prompts_group = QGroupBox(_("Prompts"))
        prompts_label = QLabel(_(
            "Modify how input and output prompts are shown in the console."
        ))
        prompts_label.setWordWrap(True)
        in_prompt_edit = self.create_lineedit(
            _("Input prompt:"),
            'in_prompt',
            "",
            _('Default is<br>'
              '<tt>In [&lt;span class="in-prompt-number"&gt;%i&lt;/span&gt;]:</tt>'),
            alignment=Qt.Horizontal,
        )
        out_prompt_edit = self.create_lineedit(
            _("Output prompt:"),
            'out_prompt',
            "",
            _('Default is<br>'
              '<tt>Out[&lt;span class="out-prompt-number"&gt;%i&lt;/span&gt;]:</tt>'),
            alignment=Qt.Horizontal,
        )

        prompts_g_layout = QGridLayout()
        prompts_g_layout.addWidget(in_prompt_edit.label, 0, 0)
        prompts_g_layout.addWidget(in_prompt_edit.textbox, 0, 1)
        prompts_g_layout.addWidget(in_prompt_edit.help_label, 0, 2)
        prompts_g_layout.addWidget(out_prompt_edit.label, 1, 0)
        prompts_g_layout.addWidget(out_prompt_edit.textbox, 1, 1)
        prompts_g_layout.addWidget(out_prompt_edit.help_label, 1, 2)

        prompts_layout = QVBoxLayout()
        prompts_layout.addWidget(prompts_label)
        prompts_layout.addLayout(prompts_g_layout)
        prompts_group.setLayout(prompts_layout)

        # Windows adjustments
        windows_group = QGroupBox(_("Windows adjustments"))
        hide_cmd_windows = newcb(
            _("Hide command line output windows "
              "generated by the subprocess module"),
            'hide_cmd_windows',
        )
        windows_layout = QVBoxLayout()
        windows_layout.addWidget(hide_cmd_windows)
        windows_group.setLayout(windows_layout)

        # --- Tabs organization ---
        self.create_tab(
            _("Interface"),
            [display_group, confirmations_group, comp_group, output_group]
        )

        self.create_tab(
            _("Plotting"),
            [matplotlib_group, backend_group, inline_group]
        )

        self.create_tab(
            _("Startup"),
            [run_lines_group, run_file_group]
        )

        self.create_tab(
            _("Advanced"),
            [autocall_group, autoreload_group, prompts_group, windows_group]
        )

    def warn_if_large_buffer(self):
        """Warn the user if the Console buffer size is very large."""
        if "buffer_size" not in self.changed_options:
            return

        msg = None
        buffer_size = self.buffer_spin.spinbox.value()

        # >10k line buffers can make Spyder slow, see spyder-ide/spyder#19091
        if buffer_size > 10_000:
            msg = _("Buffer sizes over 10000 lines can slow down Spyder")
        elif buffer_size == -1:
            msg = _("Unlimited buffer size can slow down Spyder severely")
        if msg:
            QMessageBox.warning(self, _("Warning"), msg, QMessageBox.Ok)
